
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg);

typedef struct {
		double	x0,y0,r1,r2,ra,ri;
	} Oval;

#define NOO	10	/* number of ovals */

static Oval	O[NOO]={{ 0.0 , 0.0   ,0.92 ,0.69  , 90.0, 2.0 },
			{ 0.0 ,-0.0184,0.874,0.6624, 90.0,-0.98},
			{ 0.22, 0.0   ,0.31 ,0.11  , 72.0,-0.02},
			{-0.22, 0.0   ,0.41 ,0.16  ,108.0,-0.02},
			{ 0.0 , 0.35  ,0.25 ,0.21  , 90.0, 0.01},
			{ 0.0 , 0.1   ,0.046,0.046 ,  0.0, 0.01},
			{ 0.0 ,-0.1   ,0.046,0.046 ,  0.0, 0.01},
			{-0.08,-0.605 ,0.046,0.023 ,  0.0, 0.01},
			{ 0.0 ,-0.605 ,0.023,0.023 ,  0.0, 0.01},
			{ 0.06, -0.605,0.023,0.023 , 90.0, 0.01}};

int	main(int argc,char **argv)
{
	int	N,BPS,o,n;
	double	base,step,*X[NOO],*Y[NOO],unit,W[NOO],XY[NOO],roof;
	Cell	**cell,C;

	if (argc!=6) Error("usage : 2d_hp_tg Nxy base step BPS HP_TG.tif");

	if ((N=atoi(argv[1]))<=0) Error("bad pixel number.");

	base=atof(argv[2]);

	if ((step=atof(argv[3]))<=0.0) Error("bad value step.");

	if ((BPS=atoi(argv[4]))<=0 || BPS>sizeof(Cell)*8) Error("bad BPS.");

	if ((X[0]=(double *)malloc(sizeof(double)*NOO*N))==NULL ||
	    (Y[0]=(double *)malloc(sizeof(double)*NOO*N))==NULL ||
	    (cell=(Cell **)malloc(sizeof(Cell *)*N))==NULL ||
	    (*cell=(Cell *)malloc(sizeof(Cell)*N*N))==NULL)
	    Error("memory allocation error.");

	for (o=1; o<NOO; o++) {
	    X[o]=X[o-1]+N; Y[o]=Y[o-1]+N;
	}
	for (n=1; n<N; n++) cell[n]=cell[n-1]+N;

	unit=2.0/(double)(N-1);

	for (o=0; o<NOO; o++)
{
	double	w,d,
		x0=O[o].x0,r1=1.0/O[o].r1/O[o].r1,
		y0=O[o].y0,r2=1.0/O[o].r2/O[o].r2,
		ra=O[o].ra*M_PI/180.0,
		s1=sin(ra),c1=cos(ra),xy=(r1-r2)*2.0*s1*c1,
		s2=s1*s1,c2=c1*c1,a=r1*c2+r2*s2,b=r1*s2+r2*c2;

	W[o]=(XY[o]=xy)*x0*y0+1.0;

	for (n=0; n<N; n++) {
	    w=(double)n*unit-1.0; d=  w -x0; X[o][n]=(a*d-xy*y0)*d;
				  d=(-w)-y0; Y[o][n]=(b*d-xy*x0)*d;
	}
}
	roof=base+step*(double)(C=(Cell)((1<<BPS)-1));

	for (n=0; n<N; n++)
{
	int	x;
	double	xy,ri,w=1.0-unit*(double)n;

	for (x=0; x<N; x++) {
	    xy=(unit*(double)x-1.0)*w;

	    ri=0.0;
	    for (o=0; o<NOO; o++)
		if (X[o][x]+Y[o][n]+XY[o]*xy<=W[o]) ri+=O[o].ri;

	    cell[n][x]=(ri<base)?0:(ri>roof)?C:(Cell)((ri-base)/step+0.5);
	}
}
	StoreImageFile(argv[5],N,N,BPS,cell,NULL);

	free(*cell); free(cell); free(Y[0]); free(X[0]); return 0;
}
