
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg);

typedef struct {
		double	x0,y0,z0,l1,l2,l3,h,v;
	} Ovoid;

#define NOO	10	/* number of ovoids */

static Ovoid	O[NOO]={{ 0.00, 0.000, 0.000,0.6900,0.920,0.900,  0.0, 2.00},
			{ 0.00, 0.000, 0.000,0.6624,0.874,0.880,  0.0,-0.98},
			{-0.22, 0.000,-0.250,0.4100,0.160,0.210,108.0,-0.02},
			{ 0.22, 0.000,-0.250,0.3100,0.110,0.220, 72.0,-0.02},
			{ 0.00, 0.350,-0.250,0.2100,0.250,0.500,  0.0, 0.02},
			{ 0.00, 0.100,-0.250,0.0460,0.046,0.046,  0.0, 0.02},
			{-0.08,-0.650,-0.250,0.0460,0.023,0.020,  0.0, 0.01},
			{ 0.06,-0.650,-0.250,0.0460,0.023,0.020, 90.0, 0.01},
			{ 0.06,-0.105, 0.625,0.0560,0.040,0.100, 90.0, 0.02},
			{ 0.00, 0.100, 0.625,0.0560,0.056,0.100,  0.0,-0.02}};

#define LEN	2048

int	main(int argc,char **argv)
{
	int	N,M,m,n;
	double	**R,**W,**X,**Y,u;
	char	form[LEN];

	if (argc!=4) Error("usage : 3d_hp_sg Nrt M HP_SG/");

	if ((N=atoi(argv[1]))<=0) Error("bad pixel number.");

	if ((M=atoi(argv[2]))<=0) Error("bad rotation step.");

	if ((R=(double **)malloc(sizeof(double *)*M))==NULL ||
	    (*R=(double *)malloc(sizeof(double)*M*NOO))==NULL ||
	    (W=(double **)malloc(sizeof(double *)*M))==NULL ||
	    (*W=(double *)malloc(sizeof(double)*M*NOO))==NULL ||
	    (X=(double **)malloc(sizeof(double *)*M))==NULL ||
	    (*X=(double *)malloc(sizeof(double)*M*NOO))==NULL ||
	    (Y=(double **)malloc(sizeof(double *)*M))==NULL ||
	    (*Y=(double *)malloc(sizeof(double)*M*NOO))==NULL)
	    Error("memory allocation error.");

	for (m=1; m<M; m++) {
	    R[m]=R[m-1]+NOO; W[m]=W[m-1]+NOO; X[m]=X[m-1]+NOO; Y[m]=Y[m-1]+NOO;
	}
	u=M_PI/(double)M;

	for (m=0; m<M; m++)
{
	int	o;
	double	l1,l2,l3,h2,c2,s2,w,q,h1=(double)m*u,c1=cos(h1),s1=sin(h1);

	for (o=0; o<NOO; o++) {
	    R[m][o]=O[o].x0*c1+O[o].y0*s1;

	    l1=O[o].l1*O[o].l1; l2=O[o].l2*O[o].l2; l3=O[o].l3*O[o].l3;
	    h2=h1-O[o].h*M_PI/180.0; c2=cos(h2); s2=sin(h2);
	    W[m][o]=w=4.0*l1*l2*l3/(q=(l1*c2*c2+l2*s2*s2)*l3);
	    X[m][o]=w*l3/q;
	    Y[m][o]=w/l3;
	}
}
	u=2.0/(double)(N-1);
{
	int	len=1;

	for (n=10; n<N; n*=10) ++len;
	(void)sprintf(form,"%s/%%0%dd",argv[3],len);
}
	for (n=0; n<N; n++)
{
	char	path[LEN];
	FILE	*file;
	int	l,o;
	double	r,p,x,y,w,t=(double)n*u-1.0;

	(void)sprintf(path,form,n);
	if ((file=fopen(path,"wb"))==NULL) Error("file not open.");

	(void)fprintf(file,"%d\t%d\n",N,M);

	for (m=0; m<M; m++)
	for (l=0; l<N; l++) {
	    r=(double)l*u-1.0; p=0.0;
	    for (o=0; o<NOO; o++) {
		x=r-R[m][o];
		y=t-O[o].z0;
		if ((w=W[m][o]-X[m][o]*x*x-Y[m][o]*y*y)>=0.0)
		    p+=(O[o].v*sqrt(w));
	    }
	    (void)fwrite(&p,sizeof(double),(size_t)1,file);
	}
	(void)fclose(file);
}
	free(*Y); free(Y);
	free(*X); free(X);
	free(*W); free(W);
	free(*R); free(R); return 0;
}
