
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg);

typedef struct {
		double	x0,y0,z0,r1,r2,r3,ra,ri;
	} Ovoid;

#define NOO	10	/* number of ovoids */

static Ovoid	O[NOO]={{ 0.00, 0.000, 0.000,0.6900,0.920,0.900,  0.0, 2.00},
			{ 0.00, 0.000, 0.000,0.6624,0.874,0.880,  0.0,-0.98},
			{-0.22, 0.000,-0.250,0.4100,0.160,0.210,108.0,-0.02},
			{ 0.22, 0.000,-0.250,0.3100,0.110,0.220, 72.0,-0.02},
			{ 0.00, 0.350,-0.250,0.2100,0.250,0.500,  0.0, 0.02},
			{ 0.00, 0.100,-0.250,0.0460,0.046,0.046,  0.0, 0.02},
			{-0.08,-0.650,-0.250,0.0460,0.023,0.020,  0.0, 0.01},
			{ 0.06,-0.650,-0.250,0.0460,0.023,0.020, 90.0, 0.01},
			{ 0.06,-0.105, 0.625,0.0560,0.040,0.100, 90.0, 0.02},
			{ 0.00, 0.100, 0.625,0.0560,0.056,0.100,  0.0,-0.02}};

#define LEN	2048

int	main(int argc,char **argv)
{
	int	N,BPS,o,n;
	double	base,step,*X[NOO],*Y[NOO],*Z[NOO],unit,W[NOO],XY[NOO],roof;
	Cell	**cell,C;
	char	form[LEN];

	if (argc!=6) Error("usage : 3d_hp_tg Nxyz base step BPS HP_TG/");

	if ((N=atoi(argv[1]))<=0) Error("bad pixel number.");

	base=atof(argv[2]);

	if ((step=atof(argv[3]))<=0.0) Error("bad value step.");

	if ((BPS=atoi(argv[4]))<=0 || BPS>sizeof(Cell)*8) Error("bad BPS.");

	if ((X[0]=(double *)malloc(sizeof(double)*NOO*N))==NULL ||
	    (Y[0]=(double *)malloc(sizeof(double)*NOO*N))==NULL ||
	    (Z[0]=(double *)malloc(sizeof(double)*NOO*N))==NULL ||
	    (cell=(Cell **)malloc(sizeof(Cell *)*N))==NULL ||
	    (*cell=(Cell *)malloc(sizeof(Cell)*N*N))==NULL)
	    Error("memory allocation error.");

	for (o=1; o<NOO; o++) {
	    X[o]=X[o-1]+N; Y[o]=Y[o-1]+N; Z[o]=Z[o-1]+N;
	}
	for (n=1; n<N; n++) cell[n]=cell[n-1]+N;

	unit=2.0/(double)(N-1);

	for (o=0; o<NOO; o++)
{
	double	w,d,
		x0=O[o].x0,r1=1.0/O[o].r1/O[o].r1,
		y0=O[o].y0,r2=1.0/O[o].r2/O[o].r2,
		z0=O[o].z0,r3=1.0/O[o].r3/O[o].r3,
		ra=O[o].ra*M_PI/180.0,
		s1=sin(ra),c1=cos(ra),g=(r1-r2)*2.0*s1*c1,
		s2=s1*s1,c2=c1*c1,a=r1*c2+r2*s2,b=r1*s2+r2*c2;

	W[o]=(XY[o]=g)*x0*y0+1.0;

	for (n=0; n<N; n++) {
	    w=(double)n*unit-1.0; d=w-x0; X[o][n]=(a*d-g*y0)*d;
				  d=w-y0; Y[o][n]=(b*d-g*x0)*d;
				  d=w-z0; Z[o][n]=r3*d*d;
	}
}
	roof=base+step*(double)(C=(Cell)((1<<BPS)-1));
{
	int	len=1;

	for (n=10; n<N; n*=10) ++len;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[5],len);
}
	for (n=0; n<N; n++)
{
	Cell	**Cy,*Cx;
	int	y,x;
	double	w,xy,ri;
	char	path[LEN];

	for (Cy=cell+N-1, y=0; y<N; y++, Cy--) {
	    w=unit*(double)y-1.0;
	    for (Cx=(*Cy), x=0; x<N; x++, Cx++) {
		xy=(unit*(double)x-1.0)*w;

		ri=0.0;
		for (o=0; o<NOO; o++)
		    if (X[o][x]+Y[o][y]+Z[o][n]+XY[o]*xy<=W[o]) ri+=O[o].ri;
		
		*Cx=(ri<base)?0:(ri>roof)?C:(Cell)((ri-base)/step+0.5);
	    }
	}
	(void)sprintf(path,form,n); StoreImageFile(path,N,N,BPS,cell,NULL);
}
	free(*cell); free(cell); free(Z[0]); free(Y[0]); free(X[0]); return 0;
}
