
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define time	tick	/* to prevent error of lint */

extern void	Error(char *msg);

static void	Abort(char *path,char *msg)
{
	(void)fprintf(stderr,"%s : ",path); Error(msg);
}

static int	fGetW(FILE *file)
{
	int	lo,hi;

	return ((lo=fgetc(file))==EOF || (hi=fgetc(file))==EOF)?-1:(hi<<8)|lo;
}

typedef unsigned short	Word;

typedef struct {
		union {
			char	*path;
			Word	*data;
		} u;
		double	time;
	} Table;

#define Path	u.path
#define Data	u.data
#define Time	time

static char	*fnf="file not found.";
static int	Nh,Nv,first=1;
static double	time1,time2;

#define LEN	1024

static void	Check(char *tPath,int *N,Table **T,int di)
{
	FILE	*tFile,*iFile;
	char	line[LEN],iPath[LEN];
	double	time;

	struct List {
			Table		table;
			struct List	*next;
		} *list,*next,**ptr=(&next);

	Table	*t;

	if (*tPath=='-')
	    tFile=stdin;
	else
	    if ((tFile=fopen(tPath,"r"))==NULL) Abort(tPath,fnf);

	for (*N=0; fgets(line,LEN,tFile)!=NULL &&
		   sscanf(line,"%s %lf",iPath,&time)==2; (*N)++) {
	    if ((*ptr=list=(struct List *)malloc(sizeof(struct List)))==NULL ||
	        (list->table.Path=strdup(iPath))==NULL)
		Abort(iPath,"list memory allocation error.");

	    list->table.Time=time; ptr=(&(list->next));

	    if ((iFile=fopen(iPath,"rb"))==NULL) Abort(iPath,fnf);

	    if (fgetc(iFile)!='I' ||
	        fgetc(iFile)!='M') Abort(iPath,"bad magic number.");

	    if (fGetW(iFile)<0) Abort(iPath,"bad comment length.");

	    if (first) {
		if ((Nh=fGetW(iFile))<0 ||
		    (Nv=fGetW(iFile))<0) Abort(iPath,"bad image size.");

		time1=time2=time; first=0;
	    }
	    else {
		if (fGetW(iFile)!=Nh ||
		    fGetW(iFile)!=Nv) Abort(iPath,"image size not match.");

		if (time<time1) time1=time; else if (time>time2) time2=time;
	    }
	    (void)fseek(iFile,4L,SEEK_CUR);
	    if (fGetW(iFile)!=2) Abort(iPath,"bad image type.");

	    (void)fclose(iFile);
	}
	if (tFile!=stdin) (void)fclose(tFile);

	if (*N==0) Abort(tPath,"empty table.");

	if ((*T=t=(Table *)malloc(sizeof(Table)*((di && *N==1)?2:*N)))==NULL)
	    Abort(tPath,"table memory allocation error.");

	for (*ptr=NULL; (list=next)!=NULL; t++) {
	    *t=list->table; next=list->next; free(list);
	}
}

static int	h1,v1,h2,v2;

static Word	*Read(char *path,Word *data)
{
	int	v,h,d,n=h2-h1+1;
	FILE	*file;
	Word	*D;
	short	s=127;
	char	*c=(char *)&s,*dre="data read error.";
	long	o1=h1*sizeof(Word),o2=(Nh-1-h2)*sizeof(Word);

	if (data==NULL)
	    if ((data=(Word *)malloc(sizeof(Word)*(n*(v2-v1+1))))==NULL)
		Abort(path,"data memory allocation error.");

	if ((file=fopen(path,"rb"))==NULL) Abort(path,fnf);

	(void)fseek(file,2L,SEEK_SET);
	(void)fseek(file,(long)(60+fGetW(file)+Nh*v1*sizeof(Word)),SEEK_CUR);

	D=data;
	if (*c==127)
	    for (v=v1; v<=v2; v++, D+=n) {
		(void)fseek(file,o1,SEEK_CUR);
		if (fread(D,sizeof(Word),(size_t)n,file)!=n) Abort(path,dre);
		(void)fseek(file,o2,SEEK_CUR);
	    }
	else
	    for (v=v1; v<=v2; v++) {
		(void)fseek(file,o1,SEEK_CUR);
		for (h=h1; h<=h2; h++, D++) {
		    if ((d=fGetW(file))<0) Abort(path,dre);

		    *D=(Word)d;
		}
		(void)fseek(file,o2,SEEK_CUR);
	    }

	(void)fclose(file); free(path); return data;
}

static int	Compare(Table *t1,Table *t2)
{
	return (t1->Time>t2->Time)?1:(t1->Time<t2->Time)?-1:0;
}

int	main(int argc,char **argv)
{
	int	Nd,Ni,Nt,n,Nx,Ny,y,x;
	Table	*Td,*Ti,*Tt,*T;
	double	time,R1d,R2d,R1i,R2i,d,i,t,a;
	Word	*D,*D1d,*D2d,*D1i,*D2i,*data=NULL;
	FILE	*file;

	if (argc!=2 && argc!=4 && argc!=5 && argc!=6 && argc!=7 && argc!=9)
	Error("usage : hp2xp {d.tbl i.tbl t.tbl} {v1 v2}|{h1 v1 h2 v2} XP");

	if (argc==2 || argc==4 || argc==6) {
	    Check("-",&Nd,&Td,1);
	    Check("-",&Ni,&Ti,1);
	    Check("-",&Nt,&Tt,0);
	}
	else {
	    Check(argv[1],&Nd,&Td,1);
	    Check(argv[2],&Ni,&Ti,1);
	    Check(argv[3],&Nt,&Tt,0);
	}
	switch(argc) {
	    case 2: case 5:
		h1=v1=0; h2=Nh-1; v2=Nv-1; break;
	    case 4: case 7:
		h1=0; v1=atoi(argv[argc-3]);
		h2=Nh-1; v2=atoi(argv[argc-2]); break;
	    case 6: case 9:
		h1=atoi(argv[argc-5]); v1=atoi(argv[argc-4]);
		h2=atoi(argv[argc-3]); v2=atoi(argv[argc-2]); break;
	}
	if (h1<0 || h2>=Nh || h1>h2 ||
	    v1<0 || v2>=Nv || v1>v2) Error("bad image area.");

	for (T=Td, n=0; n<Nd; n++, T++) T->Data=Read(T->Path,(Word *)NULL);

	if (Nd==1) {
	    Td->Time=time1; T->Time=time2; T->Data=Td->Data; Nd=2;
	}
	else
	    qsort(Td,(size_t)Nd,sizeof(Table),(int (*)())Compare);

	for (T=Ti, n=0; n<Ni; n++, T++) T->Data=Read(T->Path,(Word *)NULL);

	if (Ni==1) {
	    Ti->Time=time1; T->Time=time2; T->Data=Ti->Data; Ni=2;
	}
	else
	    qsort(Ti,(size_t)Ni,sizeof(Table),(int (*)())Compare);

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL)
		Abort(argv[argc-1],"file not open.");

	(void)fprintf(file,"%d\t%d\t%d\n",Nx=h2-h1+1,Ny=v2-v1+1,Nt);

	for (n=0; n<Nt; n++) {
	    D=data=Read(Tt[n].Path,data); time=Tt[n].Time;

	    for (T=Td+Nd-2; T!=Td && T->Time>time; T--) ;
	    D1d=T->Data; D2d=T[1].Data;
	    R1d=1.0-(R2d=(time-T->Time)/(T[1].Time-T->Time));

	    for (T=Ti+Ni-2; T!=Ti && T->Time>time; T--) ;
	    D1i=T->Data; D2i=T[1].Data;
	    R1i=1.0-(R2i=(time-T->Time)/(T[1].Time-T->Time));

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) {
		d=(double)*(D1d++)*R1d+(double)*(D2d++)*R2d;
		i=(double)*(D1i++)*R1i+(double)*(D2i++)*R2i;
		t=(double)*(D++);
		a=(i>d && t>d)?log((i-d)/(t-d)):0.0;
		(void)fwrite(&a,sizeof(double),(size_t)1,file);
	    }
	}
	if (file!=stdout) (void)fclose(file);

	return 0;
}
