
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg);

typedef struct {
		double	x0,y0,z0,L1,L2,L3,t0,v,	/* constants */
			t,I1,I2,I3,r1,s1,z1,z2,
			x1,y1,x2,y2,r2,s2,c;
	} Ovoid;

#define NOO	10	/* number of ovoids */

static Ovoid	O[NOO]={{ 0.00, 0.000, 0.000,0.6900,0.920,0.900,  0.0, 2.00},
			{ 0.00, 0.000, 0.000,0.6624,0.874,0.880,  0.0,-0.98},
			{-0.22, 0.000,-0.250,0.4100,0.160,0.210,108.0,-0.02},
			{ 0.22, 0.000,-0.250,0.3100,0.110,0.220, 72.0,-0.02},
			{ 0.00, 0.350,-0.250,0.2100,0.250,0.500,  0.0, 0.02},
			{ 0.00, 0.100,-0.250,0.0460,0.046,0.046,  0.0, 0.02},
			{-0.08,-0.650,-0.250,0.0460,0.023,0.020,  0.0, 0.01},
			{ 0.06,-0.650,-0.250,0.0460,0.023,0.020, 90.0, 0.01},
			{ 0.06,-0.105, 0.625,0.0560,0.040,0.100, 90.0, 0.02},
			{ 0.00, 0.100, 0.625,0.0560,0.056,0.100,  0.0,-0.02}};

int	main(int argc,char **argv)
{
	int	Nu,Nw,Nt,o,j,l,m;
	double	A,B,z,Du,Ou,Dw,Ow,t0,*U,*W,**V,y,x,t,w,u,p,a,b,d;
	FILE	*file;

	if (argc!=8 && argc!=12)
	    Error("usage : hp_xp Nu Nw Nt A B {Du Ou Dw Ow} t0 XP");

	if ((Nu=atoi(argv[1]))<=1 ||
	    (Nw=atoi(argv[2]))<=1) Error("bad number of dectors.");

	if ((Nt=atoi(argv[3]))<=0) Error("bad number of projections.");

	if ((A=atof(argv[4]))<=1.0) Error("bad source-sample distance.");
	if ((B=atof(argv[5]))<A) Error("bad source-detector distance.");

	if (argc==8) {
	    z=B/sqrt(A*A-1.0);
	    Du=z/(Ou=(double)(Nu-1)/2.0);
	    Dw=z/(Ow=(double)(Nw-1)/2.0);
	    (void)fprintf((argv[argc-1][0]=='-')?stderr:stdout,
			  "%lf\t%lf\t%lf\t%lf\n",Du,Ou,Dw,Ow);
	}
	else {
	    if ((Du=atof(argv[6]))<=0.0 ||
		(Dw=atof(argv[8]))<=0.0) Error("bad inerval of dectors.");

	    if ((Ou=atof(argv[7]))<=0.0 || Ou>=(double)(Nu-1) ||
		(Ow=atof(argv[9]))<=0.0 || Ow>=(double)(Nw-1))
		Error("bad center position of detector array.");
	}
	t0=atof(argv[argc-2])*M_PI/180.0;

	if ((U=(double *)malloc(sizeof(double)*Nu))==NULL ||
	    (W=(double *)malloc(sizeof(double)*Nw))==NULL ||
	    (V=(double **)malloc(sizeof(double *)*Nw))==NULL ||
	    (*V=(double *)malloc(sizeof(double)*Nw*Nu))==NULL)
	    Error("memory allocation error.");

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL) Error("file not open.");

	for (o=0; o<NOO; o++) {
	    sincos(O[o].t=O[o].t0*M_PI/180.0,&y,&x);
	    O[o].r1=(O[o].x0*x+O[o].y0*y)*(O[o].I1=1.0/O[o].L1);
	    O[o].s1=(O[o].y0*x-O[o].x0*y)*(O[o].I2=1.0/O[o].L2);
	    O[o].z1=z=O[o].z0*(O[o].I3=1.0/O[o].L3);
	    O[o].z2=z*z-1.0;
	}
	for (j=0; j<Nu; j++) U[j]=Du*((double)j-Ou)/B;
	for (l=0; l<Nw; l++) W[l]=Dw*((double)l-Ow)/B;

	for (l=0; l<Nw; l++) {
	    if (l>0) V[l]=V[l-1]+Nu;

	    for (j=0; j<Nu; j++) V[l][j]=2.0*sqrt(1.0+U[j]*U[j]+W[l]*W[l]);
	}
	(void)fprintf(file,"%d\t%d\t%d\n",Nu,Nw,Nt);

	for (m=0; m<Nt; m++) {
	    t=t0+2.0*M_PI*(double)m/(double)Nt;
	    for (o=0; o<NOO; o++) {
		sincos(t-O[o].t,&y,&x);
		O[o].x1=x*O[o].I1;
		O[o].y2=y*O[o].I2;
		y=O[o].r2=A*(O[o].y1=y*O[o].I1)-O[o].r1;
		x=O[o].s2=A*(O[o].x2=x*O[o].I2)+O[o].s1;
		O[o].c=x*x+y*y+O[o].z2;
	    }
	    for (l=0; l<Nw; l++) {
		w=W[l];
		for (j=0; j<Nu; j++) {
		    u=U[j]; p=0.0;
		    for (o=0; o<NOO; o++) {
			x=u*O[o].x1-O[o].y1;
			y=u*O[o].y2+O[o].x2;
			z=w*O[o].I3;
			a=x*x+y*y+z*z;
			b=x*O[o].r2-y*O[o].s2-z*O[o].z1;
			if ((d=b*b-a*O[o].c)>0.0) p+=(O[o].v*sqrt(d)/a);
		    }
		    p*=V[l][j]; (void)fwrite(&p,sizeof(double),1,file);
		}
	    }
	}
	if (file!=stdout) (void)fclose(file);

	free(*V); free(V); free(W); free(U);

	return 0;
}
