
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg);

#ifndef	RESAMPLE
#define RESAMPLE	0
#endif

static double	**P,W;

static double	BLI(double x,double y)
{
	if (x<0.0 || x>W) return 0.0;
{
	int	x0=(int)x,
		y0=(int)y,
		x1=x0+1,
		y1=y0+1;
	double	dx=x-(double)x0,
		dy=y-(double)y0,
		p0=P[y0][x0],
		px=P[y0][x1]-p0,
		py=P[y1][x0]-p0;

	return p0+px*dx+py*dy+(P[y1][x1]-p0-px-py)*dx*dy;
}}

#define	LEN	2048

int	main(int argc,char **argv)
{
	FILE	*file;
	char	str[LEN];
	int	Nf,Mf,m,Np,Mp,n;
	double	SS,SD,If,Cf,Ip,u,Cp,Sp,Sf,t,r,x,y,p;

	if (argc!=7) Error("usage : sg_f2p SG_f SS SD interval center SG_p");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,file)==NULL ||
	    sscanf(str,"%d %d",&Nf,&Mf)!=2 || Nf<=0 || Mf<=0 ||
	    (P=(double **)malloc(sizeof(double *)*(Mf+1)))==NULL ||
	    (*P=(double *)malloc(sizeof(double)*((size_t)Mf*Nf+1)))==NULL ||
	    fread(*P,sizeof(double),(size_t)Mf*Nf,file)!=(size_t)Mf*Nf)
	    Error("data read and/or memory allocation error.");

	if (file!=stdin) (void)fclose(file);

	if ((SS=atof(argv[2]))<=0.0) Error("bad source-sample distance.");
	if ((SD=atof(argv[3]))<SS) Error("bad source-detector distance.");
	if ((If=atof(argv[4]))<=0.0) Error("bad interval of detector.");
	if ((Cf=atof(argv[5]))<=0.0) Error("bad center of detector.");

	W=(double)(Nf-1); for (m=1; m<Mf; m++) P[m]=P[m-1]+Nf; P[Mf]=P[0];

	Ip=If*SS/SD;
	u=If*Cf; Cp=SS*u/sqrt(SD*SD+u*u)/Ip;
	u=If*(W-Cf); Np=(int)ceil(SS*u/sqrt(SD*SD+u*u)/Ip+Cp);
	if ((Mp=(int)ceil((double)Np*M_PI/2.0))<(m=(Mf+1)/2)) Mp=m;

	if (argv[6][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[6],"wb"))==NULL) Error("file not open.");

	(void)fprintf((file==stdout)?stderr:stdout,"%lf\t%lf\n",Ip,Cp);

	(void)fprintf(file,"%d\t%d\n",Np,Mp);

	Sp=M_PI/(double)Mp;
	Sf=2.0*M_PI/(double)Mf;

	for (m=0; m<Mp; m++) {
	    t=Sp*(double)m;
	    for (n=0; n<Np; n++) {
		r=Ip*((double)n-Cp);

		u=SD*r/sqrt(SS*SS-r*r); x=u/If; y=atan2(u,SD);
#if	RESAMPLE==1
		p=BLI(Cf+x,fmod(t+y+2.0*M_PI,2.0*M_PI)/Sf);
#elif	RESAMPLE==2
		p=BLI(Cf-x,fmod(t-y+3.0*M_PI,2.0*M_PI)/Sf);
#elif	RESAMPLE==3
		p=BLI(Cf+x,fmod(t+y+2.0*M_PI,2.0*M_PI)/Sf)-
		  BLI(Cf-x,fmod(t-y+3.0*M_PI,2.0*M_PI)/Sf);
#else
		p=(BLI(Cf+x,fmod(t+y+2.0*M_PI,2.0*M_PI)/Sf)+
		   BLI(Cf-x,fmod(t-y+3.0*M_PI,2.0*M_PI)/Sf))/2.0;
#endif
		(void)fwrite(&p,sizeof(double),(size_t)1,file);
	    }
	}
	if (file!=stdout) (void)fclose(file);

	free(*P); free(P); return 0;
}
