
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg);

static double	**P,W;

static double	BLI(double x,double y)
{
	if (x<0.0 || x>W) return 0.0;
{
	int	x0=(int)x,
		y0=(int)y,
		x1=x0+1,
		y1=y0+1;
	double	dx=x-(double)x0,
		dy=y-(double)y0,
		p0=P[y0][x0],
		px=P[y0][x1]-p0,
		py=P[y1][x0]-p0;

	return p0+px*dx+py*dy+(P[y1][x1]-p0-px-py)*dx*dy;
}}

#define	LEN	2048

int	main(int argc,char **argv)
{
	FILE	*file;
	char	str[LEN];
	int	Np,Mp,m,n,Nf,Mf;
	double	SS,SD,Ip,Cp,r,If,Cf,Sf,Sp,l,u,t,p;

	if (argc!=7) Error("usage : sg_p2f SG_p interval center SS SD SG_f");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,file)==NULL ||
	    sscanf(str,"%d %d",&Np,&Mp)!=2 || Np<=0 || Mp<=0 ||
	    (P=(double **)malloc(sizeof(double *)*(Mp+1)))==NULL ||
	    (*P=(double *)malloc(sizeof(double)*((size_t)(Mp+1)*Np+1)))==NULL ||
	    fread(*P,sizeof(double),(size_t)Mp*Np,file)!=(size_t)Mp*Np)
	    Error("data read and/or memory allocation error.");

	if (file!=stdin) (void)fclose(file);

	if ((Ip=atof(argv[2]))<=0.0) Error("bad interval of detector.");
	if ((Cp=atof(argv[3]))<=0.0) Error("bad center of detector.");
	if ((SS=atof(argv[4]))<=0.0) Error("bad source-sample distance.");
	if ((SD=atof(argv[5]))<SS) Error("bad source-detector distance.");

	W=(double)(Np-1); for (m=1; m<=Mp; m++) P[m]=P[m-1]+Np;

	r=Cp*2.0; for (n=0; n<Np; n++) P[Mp][n]=BLI(r-(double)n,0.0);

	If=Ip*SD/SS;
	r=Ip*Cp; Cf=SD*r/sqrt(SS*SS-r*r)/If;
	r=Ip*(W-Cp); Nf=(int)ceil(SD*r/sqrt(SS*SS-r*r)/If+Cf);
	if ((Mf=(int)ceil((double)Np*M_PI))<(m=Mp*2)) Mf=m;

	if (argv[6][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[6],"wb"))==NULL) Error("file not open.");

	(void)fprintf((file==stdout)?stderr:stdout,"%lf\t%lf\n",If,Cf);

	(void)fprintf(file,"%d\t%d\n",Nf,Mf);

	Sf=2.0*M_PI/(double)Mf;
	Sp=M_PI/(double)Mp;

	for (m=0; m<Mf; m++) {
	    l=Sf*(double)m;
	    for (n=0; n<Nf; n++) {
		u=If*((double)n-Cf);

		r=SS*u/sqrt(SD*SD+u*u);
		if ((t=fmod(l-atan2(u,SD)+2.0*M_PI,2.0*M_PI))<M_PI)
		    p=BLI(Cp+r/Ip,t/Sp);
		else
		    p=BLI(Cp-r/Ip,(t-M_PI)/Sp);

		(void)fwrite(&p,sizeof(double),(size_t)1,file);
	    }
	}
	if (file!=stdout) (void)fclose(file);

	free(*P); free(P); return 0;
}
